from ._base import Clock


class MockClock(Clock):
    ticks: int

    def start(self):
        self.ticks = 0

    def tick(self):
        self.ticks += 1

    def time(self) -> int:
        return self.ticks

    def to_sleep_time(self, deadline: float) -> float:
        return deadline - self.ticks



import time as ostime
import types
import typing

from ._base import Clock

__all__ = ("run", "sleep", "time")


class WallClock(Clock):
    _time: float

    def start(self):
        self._time = ostime.time()

    def time(self) -> float:
        return self._time

    def to_sleep_time(self, deadline: float) -> float:
        return deadline - self._time

    def tick(self):
        self._time = ostime.time()


DEFAULT_CLOCK = WallClock()


def run(
    coro: typing.Coroutine[None, None, typing.T], clock: Clock = None
) -> typing.T:
    if clock is None:
        clock = DEFAULT_CLOCK

    with Clock.configure(clock):
        while True:
            try:
                coro.send(None)
            except StopIteration as exc:
                return exc.value
            else:
                clock.tick()


@types.coroutine
def sleep(seconds: float):
    if seconds < 0:
        raise ValueError(f"'seconds' must be positive (got {seconds})")

    clock = Clock.get()
    start = clock.time()
    deadline = start + seconds

    while clock.to_sleep_time(deadline) > 0:
        yield


def time() -> float:
    return Clock.get().time()



from contextlib import contextmanager

__all__ = ("Clock",)


class Clock:
    _instance: "Clock"

    def __init__(self):
        self._time: float = 0

    @classmethod
    @contextmanager
    def configure(cls, clock: "Clock"):
        cls._instance = clock
        clock.start()
        try:
            yield
        finally:
            del cls._instance

    @classmethod
    def get(cls) -> "Clock":
        try:
            return cls._instance
        except AttributeError:
            raise RuntimeError(
                "Clock hasn't started yet. "
                "Did you try to access it outside of 'run()'?"
            )

    def start(self):
        raise NotImplementedError

    def time(self) -> float:
        raise NotImplementedError

    def to_sleep_time(self, deadline: float) -> float:
        raise NotImplementedError

    def tick(self):
        raise NotImplementedError



from pathlib import Path

from setuptools import find_packages, setup

with open(str(Path(".") / "README.md"), "r", encoding="utf-8") as f:
    README = f.read()

REQUIREMENTS_CI = ["pytest", "pytest-cov", "codecov", "mypy"]
REQUIREMENTS_DEV = [*REQUIREMENTS_CI, "black", "pylint", "rope", "bumpversion"]

setup(
    name="hawaiio",
    version="0.0.3",
    author="Florimond Manca",
    author_email="florimond.manca@gmail.com",
    description="A library of shiny concurrency building blocks for Python",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/florimondmanca/hawaiio.git",
    packages=find_packages(exclude=["tests*"]),
    package_data={"hawaiio": ["py.typed"]},
    zip_safe=False,
    install_requires=[],
    extras_require={"dev": REQUIREMENTS_DEV, "ci": REQUIREMENTS_CI},
    python_requires=">=3.6",
    license="MIT",
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
)
